/**
 * Copyright 2010 OpenEngSB Division, Vienna University of Technology
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.openengsb.core.workflow.internal;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.drools.KnowledgeBase;
import org.drools.event.process.DefaultProcessEventListener;
import org.drools.event.process.ProcessCompletedEvent;
import org.drools.event.process.ProcessNodeTriggeredEvent;
import org.drools.event.process.ProcessStartedEvent;
import org.drools.event.rule.BeforeActivationFiredEvent;
import org.drools.event.rule.DefaultAgendaEventListener;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.process.NodeInstance;
import org.drools.runtime.process.ProcessInstance;
import org.drools.runtime.process.WorkflowProcessInstance;
import org.drools.runtime.rule.FactHandle;
import org.drools.workflow.instance.node.SubProcessNodeInstance;
import org.openengsb.core.common.AbstractOpenEngSBService;
import org.openengsb.core.common.BundleContextAware;
import org.openengsb.core.common.Domain;
import org.openengsb.core.common.Event;
import org.openengsb.core.common.context.ContextCurrentService;
import org.openengsb.core.common.workflow.RemoteEventProcessor;
import org.openengsb.core.common.workflow.RuleBaseException;
import org.openengsb.core.common.workflow.RuleManager;
import org.openengsb.core.common.workflow.WorkflowException;
import org.openengsb.core.common.workflow.WorkflowService;
import org.openengsb.core.common.workflow.model.InternalWorkflowEvent;
import org.openengsb.core.common.workflow.model.ProcessBag;
import org.openengsb.core.common.workflow.model.RemoteEvent;
import org.openengsb.core.common.workflow.model.RuleBaseElementId;
import org.openengsb.core.common.workflow.model.RuleBaseElementType;
import org.openengsb.core.workflow.OsgiHelper;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;

public class WorkflowServiceImpl extends AbstractOpenEngSBService implements WorkflowService, RemoteEventProcessor,
        BundleContextAware, ServiceListener {

    private static final String START_FLOW_CONSEQUENCE_LINE =
        " )\nthen\n  WorkflowHelper.startFlow(kcontext.getKnowledgeRuntime(), \"%s\");\n";

    private Log log = LogFactory.getLog(WorkflowServiceImpl.class);

    private static final String FLOW_TRIGGER_RULE_TEMPLATE_START =
        "## This rule has been autogenerated by the WorkflowService\n" + "when\n" + "  %s ( name == \"%s\"";

    private static final String FLOW_TRIGGER_RULE_TEMPLATE_EVENT_FIELD = ", %s == \"%s\"";

    private RuleManager rulemanager;
    private ContextCurrentService currentContextService;
    private BundleContext bundleContext;

    private Map<String, Object> services = new HashMap<String, Object>();

    private Map<String, StatefulKnowledgeSession> sessions = new HashMap<String, StatefulKnowledgeSession>();
    private ExecutorService executor = Executors.newCachedThreadPool();

    private long timeout = 10000;
    private Lock workflowLock = new ReentrantLock();

    @Override
    public void processEvent(Event event) throws WorkflowException {
        log.info(String.format("processing Event %s of type %s", event, event.getClass()));
        StatefulKnowledgeSession session = getSessionForCurrentContext();
        FactHandle factHandle = session.insert(event);
        workflowLock.lock();
        session.fireAllRules();
        workflowLock.unlock();

        Set<Long> processIds = retrieveRelevantProcessInstanceIds(event, session);
        if (processIds.isEmpty()) {
            for (ProcessInstance p : session.getProcessInstances()) {
                p.signalEvent(event.getType(), event);
            }
        } else {
            signalEventToProcesses(event, session, processIds);
        }

        session.retract(factHandle);
    }

    @Override
    public void processRemoteEvent(RemoteEvent event) throws WorkflowException {
        processEvent(event);
    }

    private void signalEventToProcesses(Event event, StatefulKnowledgeSession session, Set<Long> processIds) {
        for (Long pid : processIds) {
            ProcessInstance processInstance = session.getProcessInstance(pid);
            if (processInstance == null) {
                log.warn(String.format("processInstance with ID %s not found, maybe it already terminated", pid));
            } else {
                processInstance.signalEvent(event.getType(), event);
            }
        }
    }

    private Set<Long> retrieveRelevantProcessInstanceIds(Event event, StatefulKnowledgeSession session) {
        Set<Long> processIds = new HashSet<Long>();
        Long processIdFromEvent = event.getProcessId();
        if (processIdFromEvent != null) {
            processIds.add(processIdFromEvent);
            processIds.addAll(getSubFlows(session.getProcessInstance(processIdFromEvent)));
        }
        if (event instanceof InternalWorkflowEvent) {
            ProcessBag bag = ((InternalWorkflowEvent) event).getProcessBag();
            Long processIdFromBag = Long.parseLong(bag.getProcessId());
            processIds.add(processIdFromBag);
            processIds.addAll(getSubFlows(session.getProcessInstance(processIdFromBag)));
        }

        return processIds;
    }

    private Collection<Long> getSubFlows(ProcessInstance processInstance) {
        Collection<Long> result = new HashSet<Long>();
        if (processInstance == null) {
            return result;
        }
        WorkflowProcessInstance wp = (WorkflowProcessInstance) processInstance;
        for (NodeInstance n : wp.getNodeInstances()) {
            if (n instanceof SubProcessNodeInstance) {
                SubProcessNodeInstance spn = (SubProcessNodeInstance) n;
                result.add(spn.getProcessInstanceId());
            }
        }
        return result;
    }

    @Override
    public long startFlow(String processId) throws WorkflowException {
        return startFlow(processId, new HashMap<String, Object>());
    }

    @Override
    public long startFlow(String processId, Map<String, Object> parameterMap) throws WorkflowException {
        try {
            return startFlowInBackground(processId, parameterMap).get();
        } catch (InterruptedException e) {
            throw new WorkflowException(e);
        } catch (ExecutionException e) {
            throw new WorkflowException(e.getCause());
        }
    }

    @Override
    public Future<Long> startFlowInBackground(String processId) throws WorkflowException {
        return startFlowInBackground(processId, new HashMap<String, Object>());
    }

    @Override
    public Future<Long> startFlowInBackground(String processId, Map<String, Object> paramterMap)
        throws WorkflowException {
        WorkflowStarter workflowStarter = new WorkflowStarter(getSessionForCurrentContext(), processId, paramterMap);
        return executor.submit(workflowStarter);
    }

    @Override
    public void registerFlowTriggerEvent(Event event, String... flowIds) throws WorkflowException {
        String eventName = event.getName();
        String ruleName = String.format("_generated_ trigger %s on %s", Arrays.asList(flowIds), eventName);
        StringBuffer ruleCode = generateFlowTriggerRule(event, flowIds);
        log.info("adding new rule with id: " + ruleName);
        try {
            rulemanager.add(new RuleBaseElementId(RuleBaseElementType.Rule, ruleName), ruleCode.toString());
        } catch (RuleBaseException e) {
            throw new WorkflowException(e);
        }
    }

    private StringBuffer generateFlowTriggerRule(Event event, String... flowIds) throws WorkflowException {
        StringBuffer ruleCode = new StringBuffer();
        ruleCode.append(String.format(FLOW_TRIGGER_RULE_TEMPLATE_START, event.getClass().getName(), event.getName()));
        addOtherPropertyChecks(event, ruleCode);
        for (String flowId : flowIds) {
            ruleCode.append(String.format(START_FLOW_CONSEQUENCE_LINE, flowId));
        }
        return ruleCode;
    }

    private void addOtherPropertyChecks(Event event, StringBuffer ruleCode) throws WorkflowException {
        Class<? extends Event> eventClass = event.getClass();
        List<PropertyDescriptor> properties = reflectPropertiesFromEventClass(eventClass);
        for (PropertyDescriptor property : properties) {
            Method getter = property.getReadMethod();
            if (Modifier.PUBLIC != getter.getModifiers()) {
                continue;
            }
            Object propertyValue = getPropertyValue(event, getter);
            if (propertyValue == null) {
                continue;
            }
            ruleCode.append(String.format(FLOW_TRIGGER_RULE_TEMPLATE_EVENT_FIELD, property.getName(), propertyValue));
        }
    }

    private Object getPropertyValue(Event event, Method getter) throws WorkflowException {
        try {
            return getter.invoke(event);
        } catch (Exception e) {
            throw new WorkflowException("Cannot invoke getter '" + getter + "' of event class '" + event.getClass()
                    + "'.", e);
        }
    }

    private List<PropertyDescriptor> reflectPropertiesFromEventClass(Class<? extends Event> clazz)
        throws WorkflowException {
        if (clazz.equals(Event.class)) {
            return new ArrayList<PropertyDescriptor>();
        }
        try {
            List<PropertyDescriptor> result = new ArrayList<PropertyDescriptor>();
            BeanInfo info = Introspector.getBeanInfo(clazz);
            result.addAll(Arrays.asList(info.getPropertyDescriptors()));

            BeanInfo eventInfo = Introspector.getBeanInfo(Event.class);
            result.removeAll(Arrays.asList(eventInfo.getPropertyDescriptors()));

            return result;
        } catch (IntrospectionException ie) {
            throw new WorkflowException("Cannot introspect event class " + clazz, ie);
        }
    }

    public void waitForFlowToFinish(long id) throws InterruptedException, WorkflowException {
        StatefulKnowledgeSession session = getSessionForCurrentContext();
        synchronized (session) {
            while (session.getProcessInstance(id) != null) {
                session.wait();
            }
        }
    }

    public void waitForFlowToFinish(long id, long timeout) throws InterruptedException, WorkflowException {
        StatefulKnowledgeSession session = getSessionForCurrentContext();
        synchronized (session) {
            while (session.getProcessInstance(id) != null) {
                session.wait(timeout);
            }
        }
    }

    public Collection<Long> getRunningFlows() throws WorkflowException {
        Collection<ProcessInstance> processInstances = getSessionForCurrentContext().getProcessInstances();
        Collection<Long> result = new HashSet<Long>();
        for (ProcessInstance p : processInstances) {
            result.add(p.getId());

        }
        return result;
    }

    private StatefulKnowledgeSession getSessionForCurrentContext() throws WorkflowException {
        String currentContextId = currentContextService.getThreadLocalContext();
        if (currentContextId == null) {
            throw new IllegalStateException("contextID must not be null");
        }
        if (sessions.containsKey(currentContextId)) {
            return sessions.get(currentContextId);
        }
        StatefulKnowledgeSession session;
        try {
            session = createSession();
        } catch (RuleBaseException e) {
            throw new WorkflowException(e);
        }
        sessions.put(currentContextId, session);
        return session;
    }

    private Collection<String> findMissingGlobals() {
        Collection<String> globalsToProcess = new ArrayList<String>(rulemanager.listGlobals().keySet());
        globalsToProcess.remove("flowHelper");
        globalsToProcess.removeAll(services.keySet());
        return discoverNewGlobalValues(globalsToProcess);
    }

    private Collection<String> discoverNewGlobalValues(Collection<String> globalsToProcess) {
        for (Iterator<String> iterator = globalsToProcess.iterator(); iterator.hasNext();) {
            String g = iterator.next();
            if (findGlobal(g)) {
                iterator.remove();
            }
        }
        return globalsToProcess;
    }

    private boolean findGlobal(String name) {
        return findDomainGlobal(name) || findNonDomainGlobal(name);
    }

    private boolean findDomainGlobal(String name) {
        String clazz = Domain.class.getName();
        String filter = String.format("(&(openengsb.service.type=domain)(id=domain.%s))", name);
        ServiceReference ref = findGlobalReference(name, clazz, filter);
        if (ref == null) {
            return false;
        }
        Object service = bundleContext.getService(ref);
        services.put(name, service);
        return true;
    }

    private ServiceReference findGlobalReference(String name, String clazz, String filter) {
        ServiceReference[] allServiceReferences;
        try {
            allServiceReferences = bundleContext.getAllServiceReferences(clazz, filter);
        } catch (InvalidSyntaxException e) {
            throw new IllegalStateException(e);
        }
        if (allServiceReferences == null) {
            return null;
        }
        if (allServiceReferences.length != 1) {
            throw new IllegalStateException(String.format("found more than one match for \"%s\".", name));
        }
        return allServiceReferences[0];
    }

    private boolean findNonDomainGlobal(String name) {
        String clazz = rulemanager.listGlobals().get(name);
        String filter =
            String.format("(&(openengsb.service.type=workflow-service)(openengsb.workflow.globalid=%s))", name);
        ServiceReference ref = findGlobalReference(name, clazz, filter);
        if (ref == null) {
            return false;
        }
        Object service = bundleContext.getService(ref);
        services.put(name, service);
        return true;
    }

    protected StatefulKnowledgeSession createSession() throws RuleBaseException, WorkflowException {
        KnowledgeBase rb = rulemanager.getRulebase();
        log.debug("retrieved rulebase: " + rb + "from source " + rulemanager);
        final StatefulKnowledgeSession session = rb.newStatefulKnowledgeSession();
        log.debug("session started");
        populateGlobals(session);
        log.debug("globals have been set");
        session.addEventListener(new DefaultProcessEventListener() {
            @Override
            public void afterProcessCompleted(ProcessCompletedEvent event) {
                synchronized (session) {
                    session.notifyAll();
                }
            }
        });
        session.addEventListener(new DefaultProcessEventListener() {
            @Override
            public void afterProcessStarted(ProcessStartedEvent event) {
                String processId2 = event.getProcessInstance().getProcessId();
                long id = event.getProcessInstance().getId();
                log.info(String.format("started process \"%s\". instance-ID: %d", processId2, id));
            }

            @Override
            public void beforeNodeTriggered(ProcessNodeTriggeredEvent event) {
                long nodeId = event.getNodeInstance().getNodeId();
                String nodeName = event.getNodeInstance().getNodeName();
                log.info(String.format("Now triggering node \"%s\" (\"%s\").", nodeName, nodeId));
            }

            @Override
            public void afterProcessCompleted(ProcessCompletedEvent event) {
                String processId2 = event.getProcessInstance().getProcessId();
                long id = event.getProcessInstance().getId();
                log.info(String.format("process completed \"%s\". instance-ID: %d", processId2, id));
            }
        });

        session.addEventListener(new DefaultAgendaEventListener() {
            @Override
            public void beforeActivationFired(BeforeActivationFiredEvent event) {
                String ruleName = event.getActivation().getRule().getName();
                log.info(String.format("rule \"%s\" fired.", ruleName));
            }
        });
        return session;
    }

    private void populateGlobals(StatefulKnowledgeSession session) throws WorkflowException {
        Collection<String> missingGlobals = findMissingGlobals();
        log.info("populating globals");
        if (missingGlobals.contains("osgiHelper")) {
            OsgiHelper osgiHelper = new OsgiHelper();
            osgiHelper.setBundleContext(bundleContext);
            session.setGlobal("osgiHelper", osgiHelper);
            missingGlobals.remove("osgiHelper");
        }
        if (!missingGlobals.isEmpty()) {
            log.info("there are still some globals missing. waiting for: ");
            log.info(missingGlobals);
            waitForGlobals(missingGlobals);
            if (!missingGlobals.isEmpty()) {
                throw new WorkflowException("there are unassigned globals, maybe some service is missing "
                        + missingGlobals);
            }
        }
        for (Entry<String, Object> entry : services.entrySet()) {
            session.setGlobal(entry.getKey(), entry.getValue());
        }
    }

    private void waitForGlobals(Collection<String> missingGlobals) throws WorkflowException {
        boolean hasChanged = true;
        while (hasChanged && !missingGlobals.isEmpty()) {
            try {
                synchronized (services) {
                    services.wait(timeout);
                }
            } catch (InterruptedException e) {
                throw new WorkflowException(e);
            }
            hasChanged = missingGlobals.removeAll(services.keySet());
        }
    }

    @Override
    public void serviceChanged(ServiceEvent event) {
        if (event.getType() == ServiceEvent.REGISTERED) {
            ServiceReference serviceReference = event.getServiceReference();
            if (serviceReference.getProperty("openengsb.service.type").equals("domain")) {
                String id = (String) serviceReference.getProperty("id");
                String name = id.replaceFirst("domain.", "");
                Domain service = (Domain) bundleContext.getService(serviceReference);
                synchronized (services) {
                    services.put(name, service);
                    services.notify();
                }
            } else if (serviceReference.getProperty("openengsb.service.type").equals("workflow-service")) {
                String name = (String) serviceReference.getProperty("openengsb.workflow.globalid");
                Object service = bundleContext.getService(serviceReference);
                synchronized (services) {
                    services.put(name, service);
                    services.notify();
                }
            }
        }

    }

    public void setCurrentContextService(ContextCurrentService currentContextService) {
        this.currentContextService = currentContextService;
    }

    public void setServices(Map<String, Object> services) {
        this.services = services;
    }

    @Override
    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    public void setRulemanager(RuleManager rulemanager) {
        this.rulemanager = rulemanager;
    }

    public void setTimeout(long timeout) {
        this.timeout = timeout;
    }

    @Override
    public void cancelFlow(Long processInstanceId) throws WorkflowException {
        getSessionForCurrentContext().abortProcessInstance(processInstanceId);
    }

}
